# -*- coding: utf-8 -*-

"""
This is for test:

this version support request header not support response header
"""
from __future__ import absolute_import

import os.path
import time
import copy
import contextlib
import threading
import uuid

from thriftpy2.thrift import TClient, TApplicationException, TMessageType, \
    TProcessor, TType
from thriftpy2.parser import load

track_method = "__thriftpy_tracing_method_name__v2"
track_thrift = load(os.path.join(os.path.dirname(__file__), "tracking.thrift"))
ctx = threading.local()


class RequestInfo(object):
    def __init__(self, request_id, api, seq, client, server, status, start,
                 end, annotation, meta):
        """Used to store call info.

        :request_id: used to identity a request
        :api: api name
        :seq: sequence number
        :client: client name
        :server: server name
        :status: request status
        :start: start timestamp
        :end: end timestamp
        :annotation: application-level key-value data
        """
        self.request_id = request_id
        self.api = api
        self.seq = seq
        self.client = client
        self.server = server
        self.status = status
        self.start = start
        self.end = end
        self.annotation = annotation
        self.meta = meta


class TTrackedClient(TClient):
    def __init__(self, tracker_handler, *args, **kwargs):
        super(TTrackedClient, self).__init__(*args, **kwargs)

        self.tracker = tracker_handler
        self._upgraded = False

        try:
            self._negotiation()
            self._upgraded = True
        except TApplicationException as e:
            if e.type != TApplicationException.UNKNOWN_METHOD:
                raise

    def _negotiation(self):
        self._oprot.write_message_begin(track_method, TMessageType.CALL,
                                        self._seqid)
        args = track_thrift.UpgradeArgs()
        self.tracker.init_handshake_info(args)
        args.write(self._oprot)
        self._oprot.write_message_end()
        self._oprot.trans.flush()

        api, msg_type, seqid = self._iprot.read_message_begin()
        if msg_type == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(self._iprot)
            self._iprot.read_message_end()
            raise x
        else:
            result = track_thrift.UpgradeReply()
            result.read(self._iprot)
            self._iprot.read_message_end()

    def _send(self, _api, **kwargs):
        if self._upgraded:
            self._header = track_thrift.RequestHeader()
            self.tracker.gen_header(self._header)
            self._header.write(self._oprot)

        self.send_start = int(time.time() * 1000)
        super(TTrackedClient, self)._send(_api, **kwargs)

    def _req(self, _api, *args, **kwargs):
        if not self._upgraded:
            return super(TTrackedClient, self)._req(_api, *args, **kwargs)

        exception = None
        status = False

        try:
            res = super(TTrackedClient, self)._req(_api, *args, **kwargs)
            status = True
            return res
        except BaseException as e:
            exception = e
            raise
        finally:
            header_info = RequestInfo(
                request_id=self._header.request_id,
                seq=self._header.seq,
                client=self.tracker.client,
                server=self.tracker.server,
                api=_api,
                status=status,
                start=self.send_start,
                end=int(time.time() * 1000),
                annotation=self.tracker.annotation,
                meta=self._header.meta,
            )
            self.tracker.record(header_info, exception)


class TTrackedProcessor(TProcessor):
    def __init__(self, tracker_handler, *args, **kwargs):
        super(TTrackedProcessor, self).__init__(*args, **kwargs)

        self.tracker = tracker_handler
        self._upgraded = False

    def process(self, iprot, oprot):
        if not self._upgraded:
            res = self._try_upgrade(iprot)
        else:
            request_header = track_thrift.RequestHeader()
            request_header.read(iprot)
            self.tracker.handle(request_header)
            res = super(TTrackedProcessor, self).process_in(iprot)

        self._do_process(iprot, oprot, *res)

    def _try_upgrade(self, iprot):
        api, msg_type, seqid = iprot.read_message_begin()
        if msg_type == TMessageType.CALL and api == track_method:
            self._upgraded = True

            args = track_thrift.UpgradeArgs()
            args.read(iprot)
            self.tracker.handle_handshake_info(args)
            result = track_thrift.UpgradeReply()
            result.oneway = False

            def call():
                pass

            iprot.read_message_end()
        else:
            result, call = self._process_in(api, iprot)

        return api, seqid, result, call

    def _process_in(self, api, iprot):
        if api not in self._service.thrift_services:
            iprot.skip(TType.STRUCT)
            iprot.read_message_end()
            return TApplicationException(
                TApplicationException.UNKNOWN_METHOD), None

        args = getattr(self._service, api + "_args")()
        args.read(iprot)
        iprot.read_message_end()
        result = getattr(self._service, api + "_result")()

        # convert kwargs to args
        api_args = [args.thrift_spec[k][1]
                    for k in sorted(args.thrift_spec)]

        def call():
            return getattr(self._handler, api)(
                *(args.__dict__[k] for k in api_args)
            )

        return result, call

    def _do_process(self, iprot, oprot, api, seqid, result, call):
        if isinstance(result, TApplicationException):
            return self.send_exception(oprot, api, result, seqid)

        try:
            result.success = call()
        except Exception as e:
            # raise if api don't have throws
            if not self.handle_exception(e, result):
                raise

        if not result.oneway:
            self.send_result(oprot, api, result, seqid)


class TrackerBase(object):
    def __init__(self, client=None, server=None):
        self.client = client
        self.server = server

    def handle(self, header):
        ctx.header = header
        ctx.counter = 0

    def gen_header(self, header):
        header.request_id = self.get_request_id()

        if not hasattr(ctx, "counter"):
            ctx.counter = 0

        ctx.counter += 1

        if hasattr(ctx, "header"):
            header.seq = "{prev_seq}.{cur_counter}".format(
                prev_seq=ctx.header.seq, cur_counter=ctx.counter)
            header.meta = ctx.header.meta
        else:
            header.meta = {}
            header.seq = str(ctx.counter)

        if hasattr(ctx, "meta"):
            header.meta.update(ctx.meta)

    def record(self, header, exception):
        pass

    @classmethod
    @contextlib.contextmanager
    def counter(cls, init=0):
        """Context for manually setting counter of seq number.

        :init: init value
        """
        if not hasattr(ctx, "counter"):
            ctx.counter = 0

        old = ctx.counter
        ctx.counter = init

        try:
            yield
        finally:
            ctx.counter = old

    @classmethod
    @contextlib.contextmanager
    def annotate(cls, **kwargs):
        ctx.annotation = kwargs
        try:
            yield ctx.annotation
        finally:
            del ctx.annotation

    @classmethod
    @contextlib.contextmanager
    def add_meta(cls, **kwds):
        if hasattr(ctx, 'meta'):
            old_dict = copy.copy(ctx.meta)
            ctx.meta.update(kwds)
            try:
                yield ctx.meta
            finally:
                ctx.meta = old_dict
        else:
            ctx.meta = kwds
            try:
                yield ctx.meta
            finally:
                del ctx.meta

    @property
    def meta(self):
        meta = ctx.header.meta if hasattr(ctx, "header") else {}
        if hasattr(ctx, "meta"):
            meta.update(ctx.meta)
        return meta

    @property
    def annotation(self):
        return ctx.annotation if hasattr(ctx, "annotation") else {}

    def get_request_id(self):
        if hasattr(ctx, "header"):
            return ctx.header.request_id
        return str(uuid.uuid4())

    def init_handshake_info(self, handshake_obj):
        pass

    def handle_handshake_info(self, handshake_obj):
        pass


class ConsoleTracker(TrackerBase):
    def record(self, header, exception):
        print(header)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import multiprocessing
import socket
import sys
import time
import uuid

import pytest

import thriftpy2

thriftpy2.install_import_hook()  # noqa

from thriftpy2.http import make_server, make_client, client_context, THttpHeaderFactory  # noqa
from thriftpy2.thrift import TApplicationException  # noqa

addressbook = thriftpy2.load(os.path.join(os.path.dirname(__file__),
                                          "addressbook.thrift"))


if sys.platform == "win32":
    pytest.skip("requires fork", allow_module_level=True)


class Dispatcher():
    def __init__(self):
        self.ab = addressbook.AddressBook()
        self.ab.people = {}

    def ping(self):
        return True

    def hello(self, name):
        return "hello " + name

    def add(self, person):
        self.ab.people[person.name] = person
        return True

    def remove(self, name):
        try:
            self.ab.people.pop(name)
            return True
        except KeyError:
            raise addressbook.PersonNotExistsError(
                "{} not exists".format(name))

    def get(self, name):
        try:
            return self.ab.people[name]
        except KeyError:
            raise addressbook.PersonNotExistsError(
                "{} not exists".format(name))

    def book(self):
        return self.ab

    def get_phonenumbers(self, name, count):
        p = [self.ab.people[name].phones[0]] if name in self.ab.people else []
        return p * count

    def get_phones(self, name):
        phone_numbers = self.ab.people[name].phones
        return dict((p.type, p.number) for p in phone_numbers)

    def sleep(self, ms):
        time.sleep(ms / 1000.0)
        return True


class CustomHeaderFactory(THttpHeaderFactory):
    def get_headers(self):
        return {"X-REQUEST-ID": str(uuid.uuid4())}


@pytest.fixture(scope="module")
def server(request):
    server = make_server(addressbook.AddressBookService, Dispatcher(),
                         host="127.0.0.1", port=6080)
    ps = multiprocessing.Process(target=server.serve)
    ps.start()

    time.sleep(0.1)

    def fin():
        if ps.is_alive():
            ps.terminate()
    request.addfinalizer(fin)


@pytest.fixture(scope="module")
def person():
    phone1 = addressbook.PhoneNumber()
    phone1.type = addressbook.PhoneType.MOBILE
    phone1.number = '555-1212'
    phone2 = addressbook.PhoneNumber()
    phone2.type = addressbook.PhoneType.HOME
    phone2.number = '555-1234'

    # empty struct
    phone3 = addressbook.PhoneNumber()

    alice = addressbook.Person()
    alice.name = "Alice"
    alice.phones = [phone1, phone2, phone3]
    alice.created_at = int(time.time())

    return alice


def client(timeout=3000):
    return client_context(addressbook.AddressBookService,
                          host="127.0.0.1", port=6080, timeout=timeout)


def client_context_with_url(timeout=3000):
    return client_context(addressbook.AddressBookService,
                          url="http://127.0.0.1:6080", timeout=timeout)


def client_context_with_malformed_path(timeout=3000):
    return client_context(addressbook.AddressBookService, host="127.0.0.1",
                          port=6080, path="foo", timeout=timeout)


def client_with_url(timeout=3000):
    return make_client(addressbook.AddressBookService,
                       url="http://127.0.0.1:6080", timeout=timeout)


def client_without_url(timeout=3000):
    return make_client(addressbook.AddressBookService, host="127.0.0.1",
                       port=6080, path="/foo", timeout=timeout)


@pytest.fixture
def client_with_malformed_path(timeout=3000):
    return make_client(addressbook.AddressBookService, host="127.0.0.1",
                       port=6080, path="foo", timeout=timeout)


def client_context_with_header_factory(timeout=3000):
    return client_context(addressbook.AddressBookService,
                          url="http://127.0.0.1:6080",
                          timeout=timeout,
                          http_header_factory=THttpHeaderFactory())


def client_context_with_custom_header_factory(timeout=3000):
    return client_context(addressbook.AddressBookService,
                          url="http://127.0.0.1:6080",
                          timeout=timeout,
                          http_header_factory=CustomHeaderFactory())


def client_with_header_factory(timeout=3000):
    return make_client(addressbook.AddressBookService,
                       url="http://127.0.0.1:6080",
                       timeout=timeout,
                       http_header_factory=THttpHeaderFactory(
                           {"X-REQUEST-ID": str(uuid.uuid4())}))


def client_with_custom_header_factory(timeout=3000):
    return make_client(addressbook.AddressBookService,
                       url="http://127.0.0.1:6080",
                       timeout=timeout,
                       http_header_factory=CustomHeaderFactory())


def test_client_context(server):
    with client() as c1, client_context_with_url() as c2,\
         client_context_with_malformed_path() as c3:
        assert c1.hello("world") == c2.hello("world") == c3.hello("world")


def test_clients(server):
    with client() as c1:
        c2 = client_with_url()
        assert c1.hello("world") == c2.hello("world")
        c2.close()


def test_clients_without_url(server):
    c = client_without_url()
    assert c.hello("world") == "hello world"


def test_client_with_malformed_path(client_with_malformed_path):
    assert client_with_malformed_path.hello("world") == "hello world"


def test_client_context_with_header_factory(server):
    with client_context_with_header_factory() as c:
        assert c.hello("world") == "hello world"


def test_client_context_custom_with_header_factory(server):
    with client_context_with_custom_header_factory() as c:
        assert c.hello("world") == "hello world"


def test_client_with_header_factory(server):
    c = client_with_header_factory()
    assert c.hello("world") == "hello world"
    c.close()


def test_client_with_custom_header_factory(server):
    c = client_with_custom_header_factory()
    assert c.hello("world") == "hello world"
    c.close()


def test_void_api(server):
    with client() as c:
        assert c.ping() is None


def test_string_api(server):
    with client() as c:
        assert c.hello("world") == "hello world"


def test_required_argument(server):
    with client() as c:
        with pytest.raises(TApplicationException):
            c.hello()

        assert c.hello(name="") == "hello "


def test_huge_res(server):
    with client() as c:
        big_str = "world" * 100000
        assert c.hello(big_str) == "hello " + big_str


def test_tstruct_req(person):
    with client() as c:
        assert c.add(person) is True


def test_tstruct_res(person):
    with client() as c:
        assert person == c.get("Alice")


def test_complex_tstruct():
    with client() as c:
        assert len(c.get_phonenumbers("Alice", 0)) == 0
        assert len(c.get_phonenumbers("Alice", 1000)) == 1000


def test_exception():
    with pytest.raises(addressbook.PersonNotExistsError):
        with client() as c:
            c.remove("Bob")


def test_client_timeout():
    with pytest.raises(socket.timeout):
        with client(timeout=200) as c:
            c.sleep(400)



from __future__ import absolute_import

import sys
import time
import traceback
from multiprocessing import Process

import pytest
import six

from thriftpy2 import _compat
from thriftpy2.thrift import TType, TPayloadMeta
import thriftpy2
from thriftpy2.http import (
    make_server as make_http_server,
    make_client as make_http_client,
)
from thriftpy2.protocol import (
    TApacheJSONProtocolFactory,
    TJSONProtocolFactory,
    TCompactProtocolFactory,
)
from thriftpy2.protocol import TBinaryProtocolFactory
from thriftpy2.rpc import make_server as make_rpc_server, \
    make_client as make_rpc_client
from thriftpy2.transport import TBufferedTransportFactory, TCyMemoryBuffer


if sys.platform == "win32":
    pytest.skip("requires fork", allow_module_level=True)


protocols = [TApacheJSONProtocolFactory,
             TJSONProtocolFactory,
             TBinaryProtocolFactory,
             TCompactProtocolFactory]


def recursive_vars(obj):
    if isinstance(obj, six.string_types):
        return six.ensure_str(obj)
    if isinstance(obj, six.binary_type):
        return six.ensure_binary(obj)
    if isinstance(obj, (int, float, bool)):
        return obj
    if isinstance(obj, dict):
        return {k: recursive_vars(v) for k, v in obj.items()}
    if isinstance(obj, (list, set)):
        return [recursive_vars(v) for v in obj]
    if hasattr(obj, '__dict__'):
        return recursive_vars(vars(obj))


@pytest.mark.parametrize('server_func',
                         [(make_rpc_server, make_rpc_client),
                          (make_http_server, make_http_client)])
@pytest.mark.parametrize('tlist', [[], ['a', 'b', 'c']])
@pytest.mark.parametrize('binary', [b'', b'\x01\x03test binary\x03\xff'])
@pytest.mark.parametrize('proto_factory', protocols)
def test_protocols(proto_factory, binary, tlist, server_func):
    test_thrift = thriftpy2.load(
        "apache_json_test.thrift",
        module_name="test_thrift"
    )
    Foo = test_thrift.Foo

    class Handler(object):
        @staticmethod
        def test(t):
            return t

    trans_factory = TBufferedTransportFactory

    def run_server():
        server = server_func[0](
            test_thrift.TestService,
            handler=Handler(),
            host='localhost',
            port=9090,
            proto_factory=proto_factory(),
            trans_factory=trans_factory(),
        )
        server.serve()

    proc = Process(target=run_server)
    proc.start()
    time.sleep(0.2)
    err = None
    try:
        test_object = test_thrift.Test(
            tdouble=12.3456,
            tint=567,
            tstr='A test \'{["string',
            tbinary=binary,
            tlist_of_strings=tlist,
            tbool=False,
            tbyte=16,
            tlong=123123123,
            tshort=123,
            tsetofints={1, 2, 3, 4, 5},
            tmap_of_int2str={
                1: "one",
                2: "two",
                3: "three"
            },
            tmap_of_str2foo={'first': Foo("first"), "2nd": Foo("baz")},
            tmap_of_str2foolist={
                'test': [Foo("test list entry")]
            },
            tmap_of_str2mapofstring2foo={
                "first": {
                    "second": Foo("testing")
                }
            },
            tmap_of_str2stringlist={
                "words": ["dog", "cat", "pie"],
                "other": ["test", "foo", "bar", "baz", "quux"]
            },
            tfoo=Foo("test food"),
            tlist_of_foo=[Foo("1"), Foo("2"), Foo("3")],
            tlist_of_maps2int=[
                {"one": 1, "two": 2, "three": 3}
            ],
            tmap_of_int2foo={
                1: Foo("One"),
                2: Foo("Two"),
                5: Foo("Five")
            },
            tbin2bin={b'Binary': b'data'},
            tset_of_binary={b'bin one', b'bin two'},
            tlist_of_binary=[b'foo roo', b'baz boo'],
        )

        client = server_func[1](
            test_thrift.TestService,
            host='localhost',
            port=9090,
            proto_factory=proto_factory(),
            trans_factory=trans_factory(),
        )
        res = client.test(test_object)
        assert recursive_vars(res) == recursive_vars(test_object)
    except Exception as e:
        traceback.print_exc()
        err = e
    finally:
        proc.terminate()
    if err:
        raise err
    time.sleep(0.1)


@pytest.mark.parametrize('server_func',
                         [(make_rpc_server, make_rpc_client),
                          (make_http_server, make_http_client)])
@pytest.mark.parametrize('proto_factory', protocols)
def test_exceptions(server_func, proto_factory):
    test_thrift = thriftpy2.load(
        "apache_json_test.thrift",
        module_name="test_thrift"
    )
    TestException = test_thrift.TestException

    class Handler(object):
        def do_error(self, arg):
            raise TestException(message=arg)

    def do_server():
        server = server_func[0](
            service=test_thrift.TestService,
            handler=Handler(),
            host='localhost',
            port=9090,
            proto_factory=proto_factory()
        )
        server.serve()

    proc = Process(target=do_server)
    proc.start()
    time.sleep(0.25)
    msg = "exception raised!"
    with pytest.raises(TestException)as e:
        client = server_func[1](
            test_thrift.TestService,
            host='localhost',
            port=9090,
            proto_factory=proto_factory()
        )
        client.do_error(msg)
    assert e.value.message == msg

    proc.terminate()
    time.sleep(1)


@pytest.mark.parametrize('proto_factory', protocols)
def test_complex_binary(proto_factory):

    spec = thriftpy2.load("bin_test.thrift", module_name="bin_thrift")
    bin_test_obj = spec.BinTest(
        tbinary=b'\x01\x0f\xffa binary string\x0f\xee',
        str2bin={
            'key': 'value',
            'foo': 'bar'
        },
        bin2bin={
            b'bin_key': b'bin_val',
            'str2bytes': b'bin bar'
        },
        bin2str={
            b'bin key': 'str val',
        },
        binlist=[b'bin one', b'bin two', 'str should become bin'],
        binset={b'val 1', b'foo', b'bar', b'baz'},
        map_of_str2binlist={
            'key1': [b'bin 1', b'pop 2']
        },
        map_of_bin2bin={
            b'abc': {
                b'def': b'val',
                b'\x1a\x04': b'\x45'
            }
        },
        list_of_bin2str=[
            {
                b'bin key': 'str val',
                b'other key\x04': 'bob'
            }
        ]
    )

    class Handler(object):
        @staticmethod
        def test(t):
            return t

    trans_factory = TBufferedTransportFactory

    def run_server():
        server = make_rpc_server(
            spec.BinService,
            handler=Handler(),
            host='localhost',
            port=9090,
            proto_factory=proto_factory(),
            trans_factory=trans_factory(),
        )
        server.serve()

    proc = Process(target=run_server)
    proc.start()
    time.sleep(0.2)

    try:
        client = make_rpc_client(
            spec.BinService,
            host='localhost',
            port=9090,
            proto_factory=proto_factory(),
            trans_factory=trans_factory(),
        )
        res = client.test(bin_test_obj)
        check_types(spec.BinTest.thrift_spec, res)
    finally:
        proc.terminate()
    time.sleep(0.2)


@pytest.mark.skipif(_compat.PYPY, reason="Must be run in cpython")
def test_complex_map():
    """
    Test from #156
    """
    from thriftpy2.protocol import cybin as proto
    b1 = TCyMemoryBuffer()
    proto.write_val(b1, TType.MAP, {"hello": "1"},
                    spec=(TType.STRING, TType.STRING))
    b1.flush()

    b2 = TCyMemoryBuffer()
    proto.write_val(b2, TType.MAP, {"hello": b"1"},
                    spec=(TType.STRING, TType.BINARY))
    b2.flush()

    assert b1.getvalue() == b2.getvalue()


type_map = {
    TType.BYTE: (int,),
    TType.I16: (int,),
    TType.I32: (int,),
    TType.I64: (int,),
    TType.DOUBLE: (float,),
    TType.STRING: six.string_types,
    TType.BOOL: (bool,),
    TType.STRUCT: TPayloadMeta,
    TType.SET: (set, list),
    TType.LIST: (list,),
    TType.MAP: (dict,),
    TType.BINARY: six.binary_type
}

type_names = {
    TType.BYTE: "Byte",
    TType.I16: "I16",
    TType.I32: "I32",
    TType.I64: "I64",
    TType.DOUBLE: "Double",
    TType.STRING: "String",
    TType.BOOL: "Bool",
    TType.STRUCT: "Struct",
    TType.SET: "Set",
    TType.LIST: "List",
    TType.MAP: "Map",
    TType.BINARY: "Binary"
}


def check_types(spec, val):
    """
    This function should check if a given thrift object matches
    a thrift spec
    Nb. This function isn't complete

    """
    if isinstance(spec, int):
        assert isinstance(val, type_map.get(spec))
    elif isinstance(spec, tuple):
        if len(spec) >= 2:
            if spec[0] in (TType.LIST, TType.SET):
                for item in val:
                    check_types(spec[1], item)
    else:
        for i in spec.values():
            t, field_name, to_type = i[:3]
            value = getattr(val, field_name)
            assert isinstance(value, type_map.get(t)), \
                "Field {} expected {} got {}".format(
                    field_name, type_names.get(t), type(value))
            if to_type:
                if t in (TType.SET, TType.LIST):
                    for _val in value:
                        check_types(to_type, _val)
                elif t == TType.MAP:
                    for _key, _val in value.items():
                        check_types(to_type[0], _key)
                        check_types(to_type[1], _val)
                elif t == TType.STRUCT:
                    check_types(to_type, value)



# -*- coding: utf-8 -*-

import thriftpy2.protocol.json as proto
from thriftpy2.protocol import TJSONProtocol
from thriftpy2.thrift import TPayload, TType
from thriftpy2.transport import TMemoryBuffer


class TItem(TPayload):
    thrift_spec = {
        1: (TType.I32, "id", False),
        2: (TType.LIST, "phones", TType.STRING, False),
    }
    default_spec = [("id", None), ("phones", None)]


def test_map_to_obj():
    val = [{"key": "ratio", "value": "0.618"}]
    spec = [TType.STRING, TType.DOUBLE]
    obj = proto.map_to_obj(val, spec)

    assert {"ratio": 0.618} == obj


def test_map_to_json():
    obj = {"ratio": 0.618}
    spec = [TType.STRING, TType.DOUBLE]
    json = proto.map_to_json(obj, spec)

    assert [{"key": "ratio", "value": 0.618}] == json


def test_list_to_obj():
    val = [4, 8, 4, 12, 67]
    spec = TType.I32
    obj = proto.list_to_obj(val, spec)

    assert [4, 8, 4, 12, 67] == obj


def test_list_to_json():
    val = [4, 8, 4, 12, 67]
    spec = TType.I32
    json = proto.list_to_json(val, spec)

    assert [4, 8, 4, 12, 67] == json


def test_struct_to_json():
    obj = TItem(id=13, phones=["5234", "12346456"])
    json = proto.struct_to_json(obj)

    assert {"id": 13, "phones": ["5234", "12346456"]} == json


def test_struct_to_obj():
    json = {"id": 13, "phones": ["5234", "12346456"]}
    obj = TItem()

    obj = proto.struct_to_obj(json, obj)

    assert obj.id == 13 and obj.phones == ["5234", "12346456"]


def test_json_proto_api_write():
    obj = TItem(id=13, phones=["5234", "12346456"])
    trans = TMemoryBuffer()

    p = TJSONProtocol(trans)
    p.write_struct(obj)

    data = trans.getvalue().decode("utf-8")
    length = data[0:4]

    import json
    data = json.loads(data[4:])

    assert length == "\x00\x00\x00S" and data == {
        "metadata": {"version": 1},
        "payload": {"phones": ["5234", "12346456"], "id": 13}}


def test_json_proto_api_read():
    obj = TItem(id=13, phones=["5234", "12346456"])
    trans = TMemoryBuffer()

    p = TJSONProtocol(trans)
    p.write_struct(obj)

    obj2 = TItem()
    obj2 = p.read_struct(obj2)

    assert obj.id == 13 and obj.phones == ["5234", "12346456"]


def test_unicode_string():
    class Foo(TPayload):
        thrift_spec = {
            1: (TType.STRING, "name", False)
        }
        default_spec = [("name", None)]

    trans = TMemoryBuffer()
    p = TJSONProtocol(trans)

    foo = Foo(name='pão de açúcar')
    foo.write(p)

    foo2 = Foo()
    foo2.read(p)

    assert foo == foo2



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
import socket
import sys
import threading
import time

from os import path
from unittest import TestCase

import pytest
from tornado import ioloop

import thriftpy2
from thriftpy2._compat import CYTHON
from thriftpy2.tornado import make_server
from thriftpy2.rpc import make_client
from thriftpy2.transport.framed import TFramedTransportFactory
from thriftpy2.protocol.binary import TBinaryProtocolFactory

try:
    import asyncio
except ImportError:
    asyncio = None


if sys.platform == "win32":
    pytest.skip("add_socket is not implemented on Windiws",
                allow_module_level=True)


logging.basicConfig(level=logging.INFO)

addressbook = thriftpy2.load(path.join(path.dirname(__file__),
                                       "addressbook.thrift"))


class Dispatcher(object):
    def __init__(self, io_loop):
        self.io_loop = io_loop
        self.registry = {}

    def add(self, person):
        """
        bool add(1: Person person);
        """
        if person.name in self.registry:
            return False
        self.registry[person.name] = person
        return True

    def get(self, name):
        """
        Person get(1: string name)
        """
        if name not in self.registry:
            raise addressbook.PersonNotExistsError()
        return self.registry[name]


class FramedTransportTestCase(TestCase):
    TRANSPORT_FACTORY = TFramedTransportFactory()
    PROTOCOL_FACTORY = TBinaryProtocolFactory()

    def mk_server(self):
        sock = self.server_sock = socket.socket(socket.AF_INET,
                                                socket.SOCK_STREAM)
        sock.bind(('127.0.0.1', 0))
        sock.setblocking(0)
        self.port = sock.getsockname()[-1]
        self.server_thread = threading.Thread(target=self.listen)
        self.server_thread.setDaemon(True)
        self.server_thread.start()

    def listen(self):
        self.server_sock.listen(128)
        if asyncio:
            # In Tornado 5.0+, the asyncio event loop will be used
            # automatically by default
            asyncio.set_event_loop(asyncio.new_event_loop())
        self.io_loop = ioloop.IOLoop.current()
        server = make_server(addressbook.AddressBookService,
                             Dispatcher(self.io_loop), io_loop=self.io_loop)
        server.add_socket(self.server_sock)
        self.io_loop.start()

    def mk_client(self):
        return make_client(addressbook.AddressBookService,
                           '127.0.0.1', self.port,
                           proto_factory=self.PROTOCOL_FACTORY,
                           trans_factory=self.TRANSPORT_FACTORY)

    def mk_client_with_url(self):
        return make_client(addressbook.AddressBookService,
                           proto_factory=self.PROTOCOL_FACTORY,
                           trans_factory=self.TRANSPORT_FACTORY,
                           url='thrift://127.0.0.1:{port}'.format(
                               port=self.port))

    def setUp(self):
        self.mk_server()
        time.sleep(0.1)
        self.client = self.mk_client()
        self.client_created_using_url = self.mk_client_with_url()

    def tearDown(self):
        self.io_loop.stop()

    def test_make_client(self):
        linus = addressbook.Person('Linus Torvalds')
        success = self.client_created_using_url.add(linus)
        assert success
        success = self.client.add(linus)
        assert not success

    def test_able_to_communicate(self):
        dennis = addressbook.Person(name='Dennis Ritchie')
        success = self.client.add(dennis)
        assert success
        success = self.client.add(dennis)
        assert not success

    def test_zero_length_string(self):
        dennis = addressbook.Person(name='')
        success = self.client.add(dennis)
        assert success
        success = self.client.get(name='')
        assert success


if CYTHON:
    from thriftpy2.transport.framed import TCyFramedTransportFactory
    from thriftpy2.protocol.cybin import TCyBinaryProtocolFactory

    class CyFramedTransportTestCase(FramedTransportTestCase):
        PROTOCOL_FACTORY = TCyBinaryProtocolFactory()
        TRANSPORT_FACTORY = TCyFramedTransportFactory()



# -*- coding: utf-8 -*-
# flake8: disable E501

'''
Tests for contrib/tracking...

Compatibility Tests Table

============= ==================================== ======================================== ========================================
client/server native                               v2                                       v3
============= ==================================== ======================================== ========================================
native        N/A                                  test_native_client_tracked_server_v2     test_native_client_tracked_server_v3
v2            test_tracked_client_v2_native_server test_tracked_client_v2_tracked_server_v2 test_tracked_client_v2_tracked_server_v3
v3            test_tracked_client_v3_native_server test_tracked_client_v3_tracked_server_v2 regular tests
============= ==================================== ======================================== ========================================
'''  # noqa

from __future__ import absolute_import

import contextlib
import multiprocessing
import os
import pickle
import random
import socket
import sys
import tempfile
import time

import thriftpy2

try:
    import dbm
except ImportError:
    import dbm.ndbm as dbm

import pytest

from thriftpy2.contrib.tracking import TTrackedProcessor, TTrackedClient, \
    TrackerBase, track_thrift
from thriftpy2.contrib.tracking.tracker import ctx

from thriftpy2.thrift import TProcessorFactory, TClient, TProcessor
from thriftpy2.server import TThreadedServer
from thriftpy2.transport import TServerSocket, TBufferedTransportFactory, \
    TTransportException, TSocket
from thriftpy2.protocol import TBinaryProtocolFactory
from compatible.version_2.tracking import (
    TTrackedProcessor as TTrackedProcessorV2,
    TTrackedClient as TTrackedClientV2,
    TrackerBase as TrackerBaseV2,
)

try:
    from pytest_cov.embed import cleanup_on_sigterm
except ImportError:
    pass
else:
    cleanup_on_sigterm()


if sys.platform == "win32":
    pytest.skip("requires fork", allow_module_level=True)


addressbook = thriftpy2.load(os.path.join(os.path.dirname(__file__),
                                          "addressbook.thrift"))
_, db_file = tempfile.mkstemp()


def _get_port():
    while True:
        port = 20000 + random.randint(1, 9999)
        for i in range(5):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex(('127.0.0.1', port))
            if result == 0:
                continue
        else:
            return port


PORT = _get_port()


class SampleTracker(TrackerBase):
    def record(self, header, exception):
        db = dbm.open(db_file, 'w')
        key = "%s:%s" % (header.request_id, header.seq)
        db[key.encode("ascii")] = pickle.dumps(header.__dict__)
        db.close()

    def handle_response_header(self, response_header):
        self.response_header = response_header

    def get_response_header(self):
        return getattr(self, 'response_header', None)


class Tracker_V2(TrackerBaseV2):
    def record(self, header, exception):
        db = dbm.open(db_file, 'w')
        key = "%s:%s" % (header.request_id, header.seq)
        db[key.encode("ascii")] = pickle.dumps(header.__dict__)
        db.close()


tracker = SampleTracker("test_client", "test_server")
tracker_v2 = Tracker_V2("test_client", "test_server")


class Dispatcher(object):
    def __init__(self):
        self.ab = addressbook.AddressBook()
        self.ab.people = {}

    def ping(self):
        test_response = {'ping': 'pong'}
        TrackerBase.add_response_meta(**test_response)
        return True

    def hello(self, name):
        # Add specially constructed response header for assertion.
        test_response = {name: name}
        TrackerBase.add_response_meta(**test_response)
        return "hello %s" % name

    def sleep(self, ms):
        return True

    def remove(self, name):
        person = addressbook.Person(name="mary")
        with client(port=PORT) as c:
            c.add(person)

        return True

    def get_phonenumbers(self, name, count):
        return [addressbook.PhoneNumber(number="sdaf"),
                addressbook.PhoneNumber(number='saf')]

    def add(self, person):
        with client(port=PORT + 1) as c:
            c.get_phonenumbers("jane", 1)

        with client(port=PORT + 1) as c:
            c.ping()
        return True

    def get(self, name):
        if not name:
            # undeclared exception
            raise ValueError('name cannot be empty')
        raise addressbook.PersonNotExistsError()


class TSampleServer(TThreadedServer):
    def __init__(self, processor_factory, trans, trans_factory, prot_factory):
        self.daemon = False
        self.processor_factory = processor_factory
        self.trans = trans

        self.itrans_factory = self.otrans_factory = trans_factory
        self.iprot_factory = self.oprot_factory = prot_factory
        self.closed = False

    def handle(self, client):
        processor = self.processor_factory.get_processor()
        itrans = self.itrans_factory.get_transport(client)
        otrans = self.otrans_factory.get_transport(client)
        iprot = self.iprot_factory.get_protocol(itrans)
        oprot = self.oprot_factory.get_protocol(otrans)
        try:
            while True:
                processor.process(iprot, oprot)
        except TTransportException:
            pass
        except Exception:
            raise
        finally:
            itrans.close()
            otrans.close()


def gen_server(port, tracker=tracker, processor=TTrackedProcessor):
    args = [processor, addressbook.AddressBookService, Dispatcher()]
    if tracker:
        args.insert(1, tracker)
    processor = TProcessorFactory(*args)
    server_socket = TServerSocket(host="localhost", port=port)
    server = TSampleServer(processor, server_socket,
                           prot_factory=TBinaryProtocolFactory(),
                           trans_factory=TBufferedTransportFactory())
    ps = multiprocessing.Process(target=server.serve)
    ps.start()
    return ps, server


@pytest.fixture(scope="module")
def server(request):
    ps, ser = gen_server(PORT)
    time.sleep(0.15)

    def fin():
        if ps.is_alive():
            ps.terminate()
            ps.join()

    request.addfinalizer(fin)
    return ser


@pytest.fixture(scope="module")
def server1(request):
    ps, ser = gen_server(PORT + 1)
    time.sleep(0.15)

    def fin():
        if ps.is_alive():
            ps.terminate()
            ps.join()

    request.addfinalizer(fin)
    return ser


@pytest.fixture(scope="module")
def server2(request):
    ps, ser = gen_server(PORT + 2)
    time.sleep(0.15)

    def fin():
        if ps.is_alive():
            ps.terminate()
            ps.join()

    request.addfinalizer(fin)
    return ser


@pytest.fixture(scope="module")
def native_server(request):
    ps, ser = gen_server(PORT + 3, tracker=None, processor=TProcessor)
    time.sleep(0.15)

    def fin():
        if ps.is_alive():
            ps.terminate()
            ps.join()

    request.addfinalizer(fin)
    return ser


@pytest.fixture(scope="module")
def tracked_server_v2(request):
    ps, ser = gen_server(PORT + 4, tracker=tracker_v2,
                         processor=TTrackedProcessorV2)
    time.sleep(0.15)

    def fin():
        if ps.is_alive():
            ps.terminate()
            ps.join()

    request.addfinalizer(fin)
    return ser


@contextlib.contextmanager
def client(client_class=TTrackedClient, port=PORT):
    socket = TSocket("localhost", port)

    try:
        trans = TBufferedTransportFactory().get_transport(socket)
        proto = TBinaryProtocolFactory().get_protocol(trans)
        trans.open()
        args = [addressbook.AddressBookService, proto]
        if client_class.__name__ == TTrackedClient.__name__:
            args.insert(0, SampleTracker("test_client", "test_server"))
        yield client_class(*args)
    finally:
        trans.close()


@pytest.fixture
def dbm_db(request):
    db = dbm.open(db_file, 'n')
    db.close()

    def fin():
        try:
            os.remove(db_file)
        except OSError:
            pass

    request.addfinalizer(fin)


@pytest.fixture
def tracker_ctx(request):
    def fin():
        if hasattr(ctx, "header"):
            del ctx.header
        if hasattr(ctx, "counter"):
            del ctx.counter
        if hasattr(ctx, "response_header"):
            del ctx.response_header

    request.addfinalizer(fin)


def test_negotiation(server):
    with client() as c:
        assert c.is_upgraded is True


def test_response_tracker(server, dbm_db, tracker_ctx):
    with client() as c:
        c.hello('you')
        assert c.tracker.response_header.meta == {'you': 'you'}
        c.hello('me')
        assert c.tracker.response_header.meta == {'me': 'me'}


def test_tracker(server, dbm_db, tracker_ctx):
    with client() as c:
        c.hello('you')
        assert c.tracker.response_header.meta == {'you': 'you'}

    time.sleep(0.2)

    db = dbm.open(db_file, 'r')
    headers = list(db.keys())
    assert len(headers) == 1

    request_id = headers[0]
    data = pickle.loads(db[request_id])

    assert "start" in data and "end" in data
    data.pop("start")
    data.pop("end")
    assert data == {
        "request_id": request_id.decode("ascii").split(':')[0],
        "seq": '1',
        "client": "test_client",
        "server": "test_server",
        "api": "hello",
        "status": True,
        "annotation": {},
        "meta": {},
    }


def test_tracker_chain(server, server1, server2, dbm_db, tracker_ctx):
    test_meta = {'test': 'test_meta'}
    with client() as c:
        with SampleTracker.add_meta(**test_meta):
            c.remove("jane")
            c.hello("yes")

    time.sleep(0.2)

    db = dbm.open(db_file, 'r')
    headers = list(db.keys())
    assert len(headers) == 5

    headers = [pickle.loads(db[i]) for i in headers]
    headers.sort(key=lambda x: x["seq"])

    assert len(set([i["request_id"] for i in headers])) == 2

    seqs = [i["seq"] for i in headers]
    metas = [i["meta"] for i in headers]
    assert seqs == ['1', '1.1', '1.1.1', '1.1.2', '2']
    assert metas == [test_meta] * 5


def test_exception(server, dbm_db, tracker_ctx):
    with pytest.raises(addressbook.PersonNotExistsError):
        with client() as c:
            c.get("jane")

    db = dbm.open(db_file, 'r')
    headers = list(db.keys())
    assert len(headers) == 1

    header = pickle.loads(db[headers[0]])
    assert header["status"] is False


def test_undeclared_exception(server, dbm_db, tracker_ctx):
    with pytest.raises(TTransportException):
        with client() as c:
            c.get('')


def test_request_id_func():
    ctx.__dict__.clear()

    header = track_thrift.RequestHeader()
    header.request_id = "hello"
    header.seq = 0

    tracker = TrackerBase()
    tracker.handle(header)

    header2 = track_thrift.RequestHeader()
    tracker.gen_header(header2)
    assert header2.request_id == "hello"


def test_annotation(server, dbm_db, tracker_ctx):
    with client() as c:
        with SampleTracker.annotate(ann="value"):
            c.ping()

        with SampleTracker.annotate() as ann:
            ann.update({"sig": "c.hello()", "user_id": "125"})
            c.hello('you')

    time.sleep(0.2)

    db = dbm.open(db_file, 'r')
    headers = list(db.keys())

    data = [pickle.loads(db[i]) for i in headers]
    data.sort(key=lambda x: x["seq"])

    assert data[0]["annotation"] == {"ann": "value"} and \
        data[1]["annotation"] == {"sig": "c.hello()", "user_id": "125"}


def test_counter(server, dbm_db, tracker_ctx):
    with client() as c:
        c.get_phonenumbers("hello", 1)

        with SampleTracker.counter():
            c.ping()
            c.hello("counter")

        c.sleep(8)

    time.sleep(0.2)

    db = dbm.open(db_file, 'r')
    headers = list(db.keys())

    data = [pickle.loads(db[i]) for i in headers]
    data.sort(key=lambda x: x["api"])
    get, hello, ping, sleep = data

    assert get["api"] == "get_phonenumbers" and get["seq"] == '1'
    assert ping["api"] == "ping" and ping["seq"] == '1'
    assert hello["api"] == "hello" and hello["seq"] == '2'
    assert sleep["api"] == "sleep" and sleep["seq"] == '2'


def test_native_client_tracked_server_v3(server):
    with client(TClient) as c:
        c.ping()
        c.hello("world")


def test_native_client_tracked_server_v2(tracked_server_v2):
    with client(TClient, port=PORT + 4) as c:
        c.ping()
        c.hello("world")


def test_tracked_client_v2_native_server(native_server):
    with client(TTrackedClientV2, PORT + 3) as c:
        assert c._upgraded is False
        c.ping()
        c.hello("cat")
        a = c.get_phonenumbers("hello", 54)
        assert len(a) == 2
        assert a[0].number == 'sdaf' and a[1].number == 'saf'


def test_tracked_client_v2_tracked_server_v2(
        tracked_server_v2, dbm_db, tracker_ctx):
    with client(TTrackedClientV2, PORT + 4) as c:
        assert c._upgraded is True

        c.ping()
        time.sleep(0.2)

        db = dbm.open(db_file, 'r')
        headers = list(db.keys())
        assert len(headers) == 1

        request_id = headers[0]
        data = pickle.loads(db[request_id])

        assert "start" in data and "end" in data
        data.pop("start")
        data.pop("end")
        assert data == {
            "request_id": request_id.decode("ascii").split(':')[0],
            "seq": '1',
            "client": "test_client",
            "server": "test_server",
            "api": "ping",
            "status": True,
            "annotation": {},
            "meta": {},
        }


def test_tracked_client_v2_tracked_server_v3(server, dbm_db, tracker_ctx):
    with client(TTrackedClientV2) as c:
        assert c._upgraded is True

        c.ping()
        time.sleep(0.2)

        db = dbm.open(db_file, 'r')
        headers = list(db.keys())
        assert len(headers) == 1

        request_id = headers[0]
        data = pickle.loads(db[request_id])

        assert "start" in data and "end" in data
        data.pop("start")
        data.pop("end")
        assert data == {
            "request_id": request_id.decode("ascii").split(':')[0],
            "seq": '1',
            "client": "test_client",
            "server": "test_server",
            "api": "ping",
            "status": True,
            "annotation": {},
            "meta": {},
        }

        assert not hasattr(ctx, 'response_header')


def test_tracked_client_v3_native_server(native_server):
    with client(port=PORT + 3) as c:
        assert c.is_upgraded is False
        c.ping()
        assert not hasattr(ctx, "response_header")

        c.hello("cat")
        a = c.get_phonenumbers("hello", 54)
        assert len(a) == 2
        assert a[0].number == 'sdaf' and a[1].number == 'saf'


def test_tracked_client_v3_tracked_server_v2(
        tracked_server_v2, dbm_db, tracker_ctx):
    with client(port=PORT + 4) as c:
        assert c.is_upgraded is True

        c.ping()
        assert not hasattr(ctx, "response_header")
        assert c.tracker.get_response_header() is None

        time.sleep(0.2)

        db = dbm.open(db_file, 'r')
        headers = list(db.keys())
        assert len(headers) == 1

        request_id = headers[0]
        data = pickle.loads(db[request_id])

        assert "start" in data and "end" in data
        data.pop("start")
        data.pop("end")
        assert data == {
            "request_id": request_id.decode("ascii").split(':')[0],
            "seq": '1',
            "client": "test_client",
            "server": "test_server",
            "api": "ping",
            "status": True,
            "annotation": {},
            "meta": {},
        }



# -*- coding: utf-8 -*-

from io import BytesIO

import pytest

from thriftpy2.protocol import compact
from thriftpy2.thrift import TPayload, TType
from thriftpy2.utils import hexlify


class TItem(TPayload):
    thrift_spec = {
        1: (TType.I32, "id", False),
        2: (TType.LIST, "phones", (TType.STRING), False),
    }
    default_spec = [("id", None), ("phones", None)]


class TPkg(TPayload):
    thrift_spec = {
        1: (TType.I32, "id", False),
        2: (TType.LIST, "items", (TType.STRUCT, TItem), False),
    }
    default_spec = [("id", None), ("items", None)]


def gen_proto(bytearray=b''):
    b = BytesIO(bytearray)
    proto = compact.TCompactProtocol(b)
    return (b, proto)


def test_pack_byte():
    b, proto = gen_proto()
    proto._write_val(TType.BYTE, 77)
    assert "4d" == hexlify(b.getvalue())


def test_unpack_byte():
    b, proto = gen_proto(b'\x4d')
    assert 77 == proto._read_val(TType.BYTE)


def test_pack_i16():
    b, proto = gen_proto()
    proto._write_val(TType.I16, 12345)
    assert "f2 c0 01" == hexlify(b.getvalue())


def test_unpack_i16():
    b, proto = gen_proto(b"\xf2\xc0\x01")
    assert 12345 == proto._read_val(TType.I16)


def test_pack_i32():
    b, proto = gen_proto()
    proto._write_val(TType.I32, 1234567890)
    assert "a4 8b b0 99 09" == hexlify(b.getvalue())


def test_unpack_i32():
    b, proto = gen_proto(b"\xa4\x8b\xb0\x99\x09")
    assert 1234567890 == proto._read_val(TType.I32)


def test_pack_i64():
    b, proto = gen_proto()
    proto._write_val(TType.I64, 1234567890123456789)
    assert "aa 84 cc de 8f bd 88 a2 22" == hexlify(b.getvalue())


def test_unpack_i64():
    b, proto = gen_proto(b"\xaa\x84\xcc\xde\x8f\xbd\x88\xa2\x22")
    assert 1234567890123456789 == proto._read_val(TType.I64)


def test_pack_double():
    b, proto = gen_proto()
    proto._write_val(TType.DOUBLE, 1234567890.1234567890)
    assert "b7 e6 87 b4 80 65 d2 41" == hexlify(b.getvalue())


def test_unpack_double():
    b, proto = gen_proto(b"\xb7\xe6\x87\xb4\x80\x65\xd2\x41")
    assert 1234567890.1234567890 == proto._read_val(TType.DOUBLE)


def test_pack_string():
    b, proto = gen_proto()
    proto._write_val(TType.STRING, "hello world!")
    assert "0c 68 65 6c 6c 6f 20 77 6f 72 6c 64 21" == \
           hexlify(b.getvalue())

    b1, proto1 = gen_proto()
    proto1._write_val(TType.STRING, "你好世界")
    assert "0c e4 bd a0 e5 a5 bd e4 b8 96 e7 95 8c" == \
           hexlify(b1.getvalue())


def test_unpack_string():
    b, proto = gen_proto(b"\x0c\x68\x65\x6c\x6c\x6f"
                         b"\x20\x77\x6f\x72\x6c\x64\x21")
    assert 'hello world!' == proto._read_val(TType.STRING)

    b, proto = gen_proto(b'\x0c\xe4\xbd\xa0\xe5\xa5'
                         b'\xbd\xe4\xb8\x96\xe7\x95\x8c')
    assert '你好世界' == proto._read_val(TType.STRING)


def test_unpack_binary():
    b, proto = gen_proto(b'\x0c\xe4\xbd\xa0\xe5\xa5'
                         b'\xbd\xe4\xb8\x96\xe7\x95\x8c')
    proto.decode_response = False

    assert '你好世界'.encode("utf-8") == proto._read_val(TType.STRING)


def test_strict_decode():
    b, proto = gen_proto(b'\x0c\xe4\xbd\xa0\xe5\xa5\x00'
                         b'\xbd\xe4\xb8\x96\xe7\x95\x8c')
    proto.strict_decode = True

    with pytest.raises(UnicodeDecodeError):
        proto._read_val(TType.STRING)


def test_pack_bool():
    b, proto = gen_proto()
    proto._write_bool(True)
    assert "01" == hexlify(b.getvalue())


def test_unpack_bool():
    b, proto = gen_proto(b"\x01")
    assert proto._read_bool()


def test_pack_container_bool():
    b, proto = gen_proto()
    proto._write_val(TType.LIST, [True, False, True], TType.BOOL)
    assert "31 01 02 01" == hexlify(b.getvalue())

    b, proto = gen_proto()
    proto._write_val(TType.MAP, {"a": True}, (TType.STRING, TType.BOOL))
    assert "01 81 01 61 01" == hexlify(b.getvalue())

    b, proto = gen_proto()
    proto._write_val(TType.MAP, {"a": [True, False]},
                     (TType.STRING, (TType.LIST, TType.BOOL)))
    assert "01 89 01 61 21 01 02" == hexlify(b.getvalue())


def test_unpack_container_bool():
    b, proto = gen_proto(b"\x31\x01\x02\x01")
    assert [True, False, True] == proto._read_val(TType.LIST, TType.BOOL)

    b, proto = gen_proto(b"\x01\x81\x01\x61\x01")
    assert {"a": True} == proto._read_val(TType.MAP, (TType.STRING, TType.BOOL))

    b, proto = gen_proto(b"\x01\x89\x01\x61\x21\x01\x02")
    assert {"a": [True, False]} == proto._read_val(
        TType.MAP, (TType.STRING, (TType.LIST, TType.BOOL)))

    b, proto = gen_proto(b"\x03\x81\x01\x61\x01\x01\x63\x01\x01\x62\x02")
    bool_hash = proto._read_val(TType.MAP, (TType.STRING, TType.BOOL))
    assert bool_hash['a'] is True
    assert bool_hash['b'] is False
    assert bool_hash['c'] is True


def test_pack_list():
    b, proto = gen_proto()
    proto._write_val(TType.LIST, [1, 2, 3, 4, 5], TType.I16)
    assert "54 02 04 06 08 0a" == hexlify(b.getvalue())


def test_unpack_list():
    b, proto = gen_proto(b"\x54\x02\x04\x06\x08\x0a")
    assert [1, 2, 3, 4, 5] == proto._read_val(TType.LIST, TType.I16)


def test_pack_map():
    b, proto = gen_proto()
    proto._write_val(TType.MAP, {'a': 2}, (TType.STRING, TType.I16))
    assert "01 84 01 61 04" == hexlify(b.getvalue())


def test_unpack_map():
    b, proto = gen_proto(b"\x01\x84\x01\x61\x04")
    assert {u'a': 2} == proto._read_val(TType.MAP, (TType.STRING, TType.I16))


def test_write_message_begin():
    b, proto = gen_proto()
    proto.write_message_begin("test", 2, 1)
    assert "82 41 01 04 74 65 73 74" == \
           hexlify(b.getvalue())


def test_read_message_begin():
    b, proto = gen_proto(b"\x82\x41\x01\x04\x74\x65\x73\x74")
    res = proto.read_message_begin()
    assert res == ("test", 2, 1)


def test_write_struct():
    b, proto = gen_proto()
    item = TItem(id=123, phones=["123456", "abcdef"])
    proto.write_struct(item)
    assert ("15 f6 01 19 28 06 31 32 33 34 "
            "35 36 06 61 62 63 64 65 66 00" == hexlify(b.getvalue()))


def test_write_struct2():
    b, proto = gen_proto()
    item = TItem(id=123, phones=["123456", "abcdef"])
    proto._write_val(TType.STRUCT, item)
    assert ("15 f6 01 19 28 06 31 32 33 34 "
            "35 36 06 61 62 63 64 65 66 00" == hexlify(b.getvalue()))


def test_read_struct():
    b, proto = gen_proto(b"\x15\xf6\x01\x19\x28\x06\x31\x32\x33\x34"
                         b"\x35\x36\x06\x61\x62\x63\x64\x65\x66\x00")
    _item = TItem(id=123, phones=["123456", "abcdef"])
    _item2 = TItem()
    proto.read_struct(_item2)
    assert _item == _item2


def test_write_struct_recur():
    b, proto = gen_proto()
    item1 = TItem(id=123, phones=["123456", "abcdef"])
    item2 = TItem(id=456, phones=["123456", "abcdef"])
    pkg = TPkg(id=123, items=[item1, item2])
    proto._write_val(TType.STRUCT, pkg)
    assert ("15 f6 01 19 2c 15 f6 01 19 28 06 31 32 33 34 35 36 06 61 62 63 "
            "64 65 66 00 15 90 07 19 28 06 31 32 33 34 35 36 06 61 62 63 64 "
            "65 66 00 00" == hexlify(b.getvalue()))


def test_read_struct_recur():
    b, proto = gen_proto(b'\x15\xf6\x01\x19,\x15\xf6\x01\x19(\x06123456\x06'
                         b'abcdef\x00\x15\x90\x07\x19(\x06123456\x06abcdef'
                         b'\x00\x00')
    pkg = TPkg()
    proto.read_struct(pkg)
    item1 = TItem(id=123, phones=["123456", "abcdef"])
    item2 = TItem(id=456, phones=["123456", "abcdef"])
    _pkg = TPkg(id=123, items=[item1, item2])
    assert _pkg == pkg


def test_write_empty_struct():
    b, proto = gen_proto()
    item = TItem()
    proto.write_struct(item)
    assert "00" == hexlify(b.getvalue())


def test_read_empty_struct():
    b, proto = gen_proto(b"\x00")
    _item = TItem()
    _item2 = TItem()
    proto.read_struct(_item2)
    assert _item == _item2


def test_write_huge_struct():
    b, proto = gen_proto()
    item = TItem(id=12345, phones=["1234567890"] * 100000)
    proto.write_struct(item)



# -*- coding: utf-8 -*-

"""
This file shows what dynamically generated
"""

from thriftpy2.thrift import (
    TPayload,
    TException,
    TType,
)


class JavaObjectArg(TPayload):
    thrift_spec = {1: (TType.I32, 'int_arg', False),
                   2: (TType.I64, 'long_arg', False),
                   3: (TType.STRING, 'string_arg', False),
                   4: (TType.BOOL, 'bool_arg', False),
                   5: (TType.BINARY, 'binary_arg', False),
                   6: (TType.DOUBLE, 'double_arg', False)}


class JavaObject(TPayload):
    thrift_spec = {1: (TType.STRING, 'full_class_name'),
                   2: (TType.LIST, 'args_list', (TType.STRUCT,
                                                 JavaObjectArg))}


class NullStruct(TPayload):
    thrift_spec = {}


class GlobalStreamId(TPayload):
    thrift_spec = {1: (TType.STRING, 'componentId'),
                   2: (TType.STRING, 'streamId')}


class Grouping(TPayload):
    thrift_spec = {1: (TType.LIST, 'fields', TType.STRING),
                   2: (TType.STRUCT, 'shuffle', NullStruct),
                   3: (TType.STRUCT, 'all', NullStruct),
                   4: (TType.STRUCT, 'none', NullStruct),
                   5: (TType.STRUCT, 'direct', NullStruct),
                   6: (TType.STRUCT, 'custom_object', JavaObject),
                   7: (TType.BINARY, 'custom_serialized'),
                   8: (TType.STRUCT, 'local_or_shuffle', NullStruct)}


class StreamInfo(TPayload):
    thrift_spec = {1: (TType.LIST, 'output_fields', TType.STRING),
                   2: (TType.BOOL, 'direct')}


class ShellComponent(TPayload):
    thrift_spec = {1: (TType.STRING, 'execution_command'),
                   2: (TType.STRING, 'script')}


class ComponentObject(TPayload):
    thrift_spec = {1: (TType.BINARY, 'serialized_java'),
                   2: (TType.STRUCT, 'shell', ShellComponent),
                   3: (TType.STRUCT, 'java_object', JavaObject)}


class ComponentCommon(TPayload):
    thrift_spec = {1: (TType.MAP, 'inputs', ((TType.STRUCT, GlobalStreamId),
                                             (TType.STRUCT, Grouping))),
                   2: (TType.MAP, 'streams', (TType.STRING, (TType.STRUCT,
                                                             StreamInfo))),
                   3: (TType.I32, 'parallelism_hint'),
                   4: (TType.STRING, 'json_conf')}


class SpoutSpec(TPayload):
    thrift_spec = {1: (TType.STRUCT, 'spout_object', ComponentObject),
                   2: (TType.STRUCT, 'common', ComponentCommon)}


class Bolt(TPayload):
    thrift_spec = {1: (TType.STRUCT, 'bolt_object', ComponentObject),
                   2: (TType.STRUCT, 'common', ComponentCommon)}


class StateSpoutSpec(TPayload):
    thrift_spec = {1: (TType.STRUCT, 'state_spout_object', ComponentObject),
                   2: (TType.STRUCT, 'common', ComponentCommon)}


class StormTopology(TPayload):
    thrift_spec = {1: (TType.MAP, 'spouts', (TType.STRING, (TType.STRUCT,
                                                            SpoutSpec))),
                   2: (TType.MAP, 'bolts', (TType.STRING, (TType.STRUCT,
                                                           Bolt))),
                   3: (TType.MAP, 'state_spouts', (TType.STRING,
                                                   (TType.STRUCT,
                                                    StateSpoutSpec)))}


class AlreadyAliveException(TException):
    thrift_spec = {1: (TType.STRING, 'msg')}


class NotAliveException(TException):
    thrift_spec = {1: (TType.STRING, 'msg')}


class InvalidTopologyException(TException):
    thrift_spec = {1: (TType.STRING, 'msg')}


class TopologySummary(TPayload):
    thrift_spec = {1: (TType.STRING, 'id'),
                   2: (TType.STRING, 'name'),
                   3: (TType.I32, 'num_tasks'),
                   4: (TType.I32, 'num_executors'),
                   5: (TType.I32, 'num_workers'),
                   6: (TType.I32, 'uptime_secs'),
                   7: (TType.STRING, 'status')}


class SupervisorSummary(TPayload):
    thrift_spec = {1: (TType.STRING, 'host'),
                   2: (TType.I32, 'uptime_secs'),
                   3: (TType.I32, 'num_workers'),
                   4: (TType.I32, 'num_used_workers'),
                   5: (TType.STRING, 'supervisor_id')}


class ClusterSummary(TPayload):
    thrift_spec = {1: (TType.LIST, 'supervisors', (TType.STRUCT,
                                                   SupervisorSummary)),
                   2: (TType.I32, 'nimbus_uptime_secs'),
                   3: (TType.LIST, 'topologies', (TType.STRUCT,
                                                  TopologySummary))}


class ErrorInfo(TPayload):
    thrift_spec = {1: (TType.STRING, 'error', False),
                   2: (TType.I32, 'error_time_secs', False)}


class BoltStats(TPayload):
    thrift_spec = {1: (TType.MAP, 'acked', (TType.STRING, (TType.MAP,
                                                           ((TType.STRUCT,
                                                             GlobalStreamId),
                                                            TType.I64)))),
                   2: (TType.MAP, 'failed', (TType.STRING, (TType.MAP,
                                                            ((TType.STRUCT,
                                                              GlobalStreamId),
                                                             TType.I64)))),
                   3: (TType.MAP, 'process_ms_avg', (TType.STRING,
                                                     (TType.MAP,
                                                      ((TType.STRUCT,
                                                        GlobalStreamId),
                                                       TType.DOUBLE)))),
                   4: (TType.MAP, 'executed', (TType.STRING,
                                               (TType.MAP, ((TType.STRUCT,
                                                             GlobalStreamId),
                                                            TType.I64)))),
                   5: (TType.MAP, 'execute_ms_avg', (TType.STRING,
                                                     (TType.MAP,
                                                      ((TType.STRUCT,
                                                        GlobalStreamId),
                                                       TType.DOUBLE))))}


class SpoutStats(TPayload):
    thrift_spec = {1: (TType.MAP, 'acked', (TType.STRING, (TType.MAP,
                                                           (TType.STRING,
                                                            TType.I64)))),
                   2: (TType.MAP, 'failed', (TType.STRING, (TType.MAP,
                                                            (TType.STRING,
                                                             TType.I64)))),
                   3: (TType.MAP, 'complete_ms_avg', (TType.STRING,
                                                      (TType.MAP,
                                                       (TType.STRING,
                                                        TType.DOUBLE))))}


class ExecutorSpecificStats(TPayload):
    thrift_spec = {1: (TType.STRUCT, 'bolt', BoltStats),
                   2: (TType.STRUCT, 'spout', SpoutStats)}


class ExecutorStats(TPayload):
    thrift_spec = {1: (TType.MAP, 'emitted', (TType.STRING, (TType.MAP,
                                                             (TType.STRING,
                                                              TType.I64)))),
                   2: (TType.MAP, 'transferred', (TType.STRING,
                                                  (TType.MAP, (TType.STRING,
                                                               TType.I64)))),
                   3: (TType.STRUCT, 'specific', ExecutorSpecificStats)}


class ExecutorInfo(TPayload):
    thrift_spec = {1: (TType.I32, 'task_start'),
                   2: (TType.I32, 'task_end')}


class ExecutorSummary(TPayload):
    thrift_spec = {1: (TType.STRUCT, 'executor_info', ExecutorInfo),
                   2: (TType.STRING, 'component_id'),
                   3: (TType.STRING, 'host'),
                   4: (TType.I32, 'port'),
                   5: (TType.I32, 'uptime_secs'),
                   7: (TType.STRUCT, 'stats', ExecutorStats)}


class TopologyInfo(TPayload):
    thrift_spec = {1: (TType.STRING, 'id'),
                   2: (TType.STRING, 'name'),
                   3: (TType.I32, 'uptime_secs'),
                   4: (TType.LIST, 'executors', (TType.STRUCT,
                                                 ExecutorSummary)),
                   5: (TType.STRING, 'status'),
                   6: (TType.MAP, 'errors', (TType.STRING, (TType.LIST,
                                                            (TType.STRUCT,
                                                             ErrorInfo))))}


class KillOptions(TPayload):
    thrift_spec = {1: (TType.I32, 'wait_secs')}


class RebalanceOptions(TPayload):
    thrift_spec = {1: (TType.I32, 'wait_secs'),
                   2: (TType.I32, 'num_workers'),
                   3: (TType.MAP, 'num_executors', (TType.STRING, TType.I32))}


class TopologyInitialStatus(object):
    ACTIVE = 1
    INACTIVE = 2


class SubmitOptions(TPayload):
    thrift_spec = {1: (TType.I32, 'initial_status')}


class Nimbus(object):
    ''' Nimbus service '''
    class submitTopology_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name'),
                       2: (TType.STRING, 'uploadedJarLocation'),
                       3: (TType.STRING, 'jsonConf'),
                       4: (TType.STRUCT, 'topology', StormTopology)}

    class submitTopology_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', AlreadyAliveException),
                       2: (TType.STRUCT, 'ite', InvalidTopologyException)}

    class submitTopologyWithOpts_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name'),
                       2: (TType.STRING, 'uploadedJarLocation'),
                       3: (TType.STRING, 'jsonConf'),
                       4: (TType.STRUCT, 'topology', StormTopology),
                       5: (TType.STRUCT, 'options', SubmitOptions)}

    class submitTopologyWithOpts_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', AlreadyAliveException),
                       2: (TType.STRUCT, 'ite', InvalidTopologyException)}

    class killTopology_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name')}

    class killTopology_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', NotAliveException)}

    class killTopologyWithOpts_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name'),
                       2: (TType.STRUCT, 'options', KillOptions)}

    class killTopologyWithOpts_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', NotAliveException)}

    class activate_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name')}

    class activate_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', NotAliveException)}

    class deactivate_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name')}

    class deactivate_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', NotAliveException)}

    class rebalance_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'name'),
                       2: (TType.STRUCT, 'options', RebalanceOptions)}

    class rebalance_result(TPayload):
        thrift_spec = {1: (TType.STRUCT, 'e', NotAliveException),
                       2: (TType.STRUCT, 'ite', InvalidTopologyException)}

    class beginFileUpload_args(TPayload):
        thrift_spec = {}

    class beginFileUpload_result(TPayload):
        thrift_spec = {0: (TType.STRING, 'success')}

    class uploadChunk_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'location'),
                       2: (TType.BINARY, 'chunk')}

    class uploadChunk_result(TPayload):
        thrift_spec = {}

    class finishFileUpload_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'location')}

    class finishFileUpload_result(TPayload):
        thrift_spec = {}

    class beginFileDownload_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'file')}

    class beginFileDownload_result(TPayload):
        thrift_spec = {0: (TType.STRING, 'success')}

    class downloadChunk_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id')}

    class downloadChunk_result(TPayload):
        thrift_spec = {0: (TType.BINARY, 'success')}

    class getNimbusConf_args(TPayload):
        thrift_spec = {}

    class getNimbusConf_result(TPayload):
        thrift_spec = {0: (TType.STRING, 'success')}

    class getClusterInfo_args(TPayload):
        thrift_spec = {}

    class getClusterInfo_result(TPayload):
        thrift_spec = {0: (TType.STRUCT, 'success', ClusterSummary)}

    class getTopologyInfo_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id')}

    class getTopologyInfo_result(TPayload):
        thrift_spec = {0: (TType.STRUCT, 'success', TopologyInfo),
                       1: (TType.STRUCT, 'e', NotAliveException)}

    class getTopologyConf_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id')}

    class getTopologyConf_result(TPayload):
        thrift_spec = {0: (TType.STRING, 'success'),
                       1: (TType.STRUCT, 'e', NotAliveException)}

    class getTopology_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id')}

    class getTopology_result(TPayload):
        thrift_spec = {0: (TType.STRUCT, 'success', StormTopology),
                       1: (TType.STRUCT, 'e', NotAliveException)}

    class getUserTopology_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id')}

    class getUserTopology_result(TPayload):
        thrift_spec = {0: (TType.STRUCT, 'success', StormTopology),
                       1: (TType.STRUCT, 'e', NotAliveException)}


class DRPCRequest(TPayload):
    thrift_spec = {1: (TType.STRING, 'func_args'),
                   2: (TType.STRING, 'request_id')}


class DRPCExecutionException(TException):
    thrift_spec = {1: (TType.STRING, 'msg')}


class DistributedRPC(object):
    ''' DistributedRPC service '''
    class execute_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'functionName'),
                       2: (TType.STRING, 'funcArgs')}

    class execute_result(TPayload):
        thrift_spec = {0: (TType.STRING, 'success'),
                       1: (TType.STRUCT, 'e', DRPCExecutionException)}


class DistributedRPCInvocations(object):
    ''' DistributedRPCInvocations service '''
    class result_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id'),
                       2: (TType.STRING, 'result')}

    class result_result(TPayload):
        thrift_spec = {}

    class fetchRequest_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'functionName')}

    class fetchRequest_result(TPayload):
        thrift_spec = {0: (TType.STRUCT, 'success', DRPCRequest)}

    class failRequest_args(TPayload):
        thrift_spec = {1: (TType.STRING, 'id')}

    class failRequest_result(TPayload):
        thrift_spec = {}
